#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

static int apply_patch(int fd, off_t offset, const char* oldcode, const char* newcode, size_t codelen) {
	char buf[4096];

	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 20;
	}
	if (read(fd, buf, codelen) != codelen) {
		return 21;
	}
	if (!memcmp(buf, newcode, codelen)) {
		return 1;
	}
	if (memcmp(buf, oldcode, codelen)) {
		return 10;
	}
	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 30;
	}
	if (write(fd, newcode, codelen) != codelen) {
		return 40;
	}
	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 41;
	}
	if (read(fd, buf, codelen) != codelen) {
		return 42;
	}
	if (!memcmp(buf, newcode, codelen)) {
		return 0;
	}
	if (memcmp(buf, oldcode, codelen)) {
		return 43;
	}
	return 44;
}

static int trypatch30(int fd, off_t offset, char* name) {
	static const char newcode1[] = 
"\x83\xF8\x04\x72\x1D\x8B\x57\x18\xC7\x42\x2C\x24\x00\x00\x00";
	static const char oldcode1[] = 
"\x83\xF8\x04\x75\x1D\x8B\x57\x18\xC7\x42\x2C\x24\x00\x00\x00";
	static const char newcode2[] = 
"\x53\x83\xEC\x58\xB8\x74\x00\x00\x00\x89\xE3\xCD\x80\x85\xC0\x75"
"\x18\x8B\x4C\x24\x14\x8B\x54\x24\x10\x8B\x44\x24\x34\x83\xF8\x01"
"\x76\x11\x3D\x00\x10\x00\x00\x74\x10\x31\xC9\x31\xC0\xEB\x2E\xDA"
"\x88\x10\x08\xC1\xEA\x0C\xC1\xE9\x0C\x31\xC0\x81\xFA\xFF\x7F\x00"
"\x00\x76\x0C\x81\xC2\xFF\x1F\x00\x00\x81\xE2\x00\xE0\xFF\xFF\x81"
"\xC2\xFF\x07\x00\x00\x81\xE2\x00\xF8\xFF\xFF\x40\xB3\x99\x8B\x5C"
"\x24\x60\xC7\x03\x80\x00\x00\x00\x8B\x5C\x24\x64\x89\x13\x8B\x5C"
"\x24\x68\x89\x0B\x83\xC4\x58\x5B\xC3";
	static const char oldcode2[] = 
"\x55\x89\xE5\x83\xEC\x0C\x57\x56\x53\x8B\x45\x08\x81\xC4\x18\xFC"
"\xFF\xFF\x89\xE3\xC7\x00\x80\x00\x00\x00\x8B\x4D\x0C\xC7\x01\x00"
"\x00\x00\x00\x8B\x7D\x10\xC7\x07\x00\x00\x00\x00\x6A\x00\x68\xDA"
"\x88\x10\x08\xE8\xC0\xD9\xF8\xFF\x89\xC6\x83\xC4\x08\x85\xF6\x7D"
"\x07\x31\xC0\xE9\x9A\x00\x00\x00\x68\xE7\x03\x00\x00\x53\x56\xE8"
"\x94\xD5\xF8\xFF\x83\xC4\x0C\x85\xC0\x7C\x37\xC6\x04\x18\x00\x8D"
"\x53\x01\x80\x3B\x0A\x74\x0C\x90\x8A\x02\x84\xC0\x74\x05\x42\x3C"
"\x0A\x75\xF5\x8D\x45\xF4\x50\x8D\x45";
	int ret;
	int ret2;

	ret = apply_patch(fd, 0x20F8C, oldcode1, newcode1, sizeof(newcode1) - 1);
	if (ret >= 10) {
		return ret;
	}
	ret2 = ret;
	ret = apply_patch(fd, 479300, oldcode2, newcode2, sizeof(newcode2) - 1);
	if (ret >= 10) {
		return ret;
	}
	ret2 += ret;
	if (ret2 == 0) {
		printf("%s, now patched\n", name);
		return 0;
	}
	if (ret2 < 2) {
		printf("%s, was partially patched, now fully patched\n", name);
		return 0;
	}
	printf("%s, already patched\n", name);
	return 0;
}

static struct vmp {
	off_t offs;
	char* name;
	int (*fn)(int fd, off_t offset, char* name);
} vmpinfo[] = {
	{ 0x20F8C, "VMware-1455", trypatch30 },
	{ 0,       NULL,         NULL }};

int main(int argc, char* argv[]) {
	int fd;
	int err;
	struct vmp *vp;
		
	if (argc < 2) {
		fprintf(stderr, "File name missing!\n");
		return 99;
	}
	printf("Updating %s ... ", argv[1]);
	fflush(stdout);
	fd = open(argv[1], O_RDWR);
	if (fd == -1) {
		printf("failed\n");
		fprintf(stderr, "Cannot open %s: %m\n", argv[1]);
		return 98;
	}
	for (vp = vmpinfo; vp->name; vp++) {
		err = vp->fn(fd, vp->offs, vp->name);
		switch (err) {
			case 0 ... 9:	close(fd); return 0;	/* OK */
			case 40 ... 49: 
				close(fd);
				fprintf(stderr, "ERROR! It is possible that your vmware binary is corrupted.\n");
				fprintf(stderr, "ERROR! Please, reinstall vmware. Error code is %u.\n", err);
				return 40;
			case 30 ... 39:
				close(fd);
				fprintf(stderr, "ERROR! Cannot update vmware binary. Error code is %u.\n", err);
				return 30;
			case 10 ... 29:
				break;
			default:
				close(fd);
				fprintf(stderr, "ERROR! Unexpected error. Error code is %u.\n", err);
				return 99;
		}
	}
	close(fd);
	printf("Unknown version\n");
	fprintf(stderr, "ERROR! Sorry, unknown version of vmware is installed on this machine.\n");
	return 10;
}

